@@ -6,7 +6,7 @@
 	Designed and implemented in July of 1991 by John Walker
 
     Release 5: August 1995.  Added GSM compression, DES, IDEA, and
-			     one-time pad encryption.
+			     key file encryption.
 
     Release 5.1: September 1995.  Added ADPCM compression, made
 			     push-to-talk mode the default, switched
@@ -26,12 +26,14 @@
 /*  Destination host descriptor.  */
 
 struct destination {
-    struct destination *dnext;
-    char *server;
-    struct sockaddr_in name;
+    struct destination *dnext;	      /* Next destination in list */
+    char *server;		      /* Host name identifier string */
+    struct sockaddr_in name;	      /* Internet address */
+    unsigned char time_to_live;       /* Multicast scope specification */
     char deskey[9];		      /* Destination DES key, if any */
     char ideakey[17];		      /* Destination IDEA key, if any */
-    char *otp;			      /* One-time pad */
+    char pgpkey[17];		      /* Destination PGP key, if any */
+    char *otp;			      /* Key file */
 };
 
 static char *progname;		      /* Program name */
@@ -57,10 +59,16 @@ static gsm gsmh;		      /* GSM handle */
 static struct adpcm_state adpcm;      /* ADPCM compression state */
 static char curkey[9] = "";           /* Current DES key if curkey[0] != 0 */
 static char curideakey[17] = "";      /* Current IDEA key if curideakey[0] != 0 */
-static char *curotp = NULL;	      /* One-time pad buffer */
+static char curpgpkey[17] = "";       /* Current PGP key if curpgpkey[0] != 0 */
+static char *curotp = NULL;	      /* Key file buffer */
 #ifdef TINY_PACKETS
 static int sound_packet;	      /* Ideal samples/packet */
 #endif
+static struct soundbuf *pgpsb = NULL; /* PGP key sound buffer, if any */
+#ifdef HALF_DUPLEX
+static struct in_addr localhost;      /* Our internet address */
+static int halfDuplexMuted = FALSE;   /* Muted by half-duplex transmission */
+#endif
 
 #ifdef sgi
 static long usticks;		      /* Microseconds per clock tick */
@@ -127,29 +135,78 @@ static int addest(host)
   char *host;
 {
     struct destination *d;
-    struct hostent *hp, *gethostbyname();
+    struct hostent *h;
+    long naddr;
+    unsigned int ttl = 1;
+    char *mcs;
 
-    hp = gethostbyname(host);
-    if (hp == 0) {
-        fprintf(stderr, "%s: unknown host\n", host);
-	return FALSE;
+    /* If a multicast scope descriptor appears in the name, scan
+       it and lop it off the end.  We'll apply it later if we discover
+       this is actually a multicast address. */
+
+    if ((mcs = strchr(host, '/')) != NULL) {
+	*mcs++ = 0;
+#ifdef MULTICAST
+	ttl = atoi(mcs);
+#endif
     }
 
+    /* If it's a valid IP number, use it.  Otherwise try to look
+       up as a host name. */
+
+    if ((naddr = inet_addr(host)) != -1) {
+    } else {
+	h = gethostbyname(host);
+	if (h == 0) {
+            fprintf(stderr, "%s: unknown host\n", host);
+	    return FALSE;
+	}
+	bcopy((char *) h->h_addr, (char *) &naddr, sizeof naddr);
+    }
+
+#ifdef MULTICAST
+    if (!IN_MULTICAST(naddr)) {
+	ttl = 0;
+    }
+#endif
+
     d = (struct destination *) malloc(sizeof(struct destination));
     d->dnext = NULL;
     d->server = host;
-    bcopy((char *) hp->h_addr, (char *) &(d->name.sin_addr), hp->h_length);
+    bcopy((char *) &naddr, (char *) &(d->name.sin_addr), sizeof naddr);
     bcopy(curkey, d->deskey, 9);
     bcopy(curideakey, d->ideakey, 17);
+    bcopy(curpgpkey, d->pgpkey, 17);
     d->otp = curotp;
+    d->time_to_live = ttl;
     d->name.sin_family = AF_INET;
-    d->name.sin_port = htons(2074);
+    d->name.sin_port = htons(Internet_Port);
     if (dests == NULL) {
 	dests = d;
     } else {
 	dtail->dnext = d;
     }
     dtail = d;
+
+    if (pgpsb != NULL) {
+	int i;
+
+        /* "What I tell you three times is true".  When we change
+	   over to a TCP connection bracketing the UDP sound data
+	   transmission, we can send this just once, knowing it has
+	   arrived safely. */
+
+	for (i = 0; i < 3; i++) {
+	    if (sendto(sock, pgpsb, (sizeof(struct soundbuf) - BUFL) +
+				    pgpsb->buffer.buffer_len,
+		0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
+                perror("sending PGP session key");
+		break;
+	    }
+	}
+/* No need to sleep once ack from far end confirms key decoded. */
+sleep(7);
+    }
     return TRUE;
 }
 
@@ -169,7 +226,17 @@ static int sendpkt(sb)
     }
 
     for (d = dests; d != NULL; d = d->dnext) {
-	if (d->deskey[0] || d->ideakey[0] || d->otp != NULL) {
+#ifdef MULTICAST
+	if (IN_MULTICAST(d->name.sin_addr.s_addr)) {
+	    setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL,
+		&(d->time_to_live), sizeof d->time_to_live);
+	    if (debugging) {
+                fprintf(stderr, "Multicasting with scope of %d to %s.\n",
+		    d->time_to_live, inet_ntoa(d->name.sin_addr));
+	    }
+	}
+#endif
+	if (d->deskey[0] || d->ideakey[0] || d->pgpkey[0] || d->otp != NULL) {
 	    soundbuf ebuf;
 	    int i;
 	    LONG slen;
@@ -230,11 +297,33 @@ static int sendpkt(sb)
 		ebuf.compression |= fEncIDEA;
 	    }
 
-	    /* One-time pad encryption. */
+	    /* PGP encryption. */
+
+	    if (d->pgpkey[0]) {
+		unsigned short iv[4];
+
+		bzero(iv, sizeof(iv));
+		initcfb_idea(iv, d->pgpkey + 1, FALSE);
+
+                /* If we're PGP IDEA encrypting we must round the size of
+		   the data to be sent to be a multiple of 8 so that
+		   the entire IDEA frame is sent. */
+
+		slen = (slen + 7) & (~7);
+		if (debugging) {
+                    fprintf(stderr, "Encrypting %d bytes with PGP key.\r\n",
+			    slen);
+		}
+		ideacfb(ebuf.buffer.buffer_val, slen);
+		close_idea();
+		ebuf.compression |= fEncPGP;
+	    }
+
+	    /* Key file encryption. */
 
 	    if (d->otp != NULL) {
 		if (debugging) {
-                    fprintf(stderr, "Encrypting %d bytes with one-time pad.\r\n",
+                    fprintf(stderr, "Encrypting %d bytes with key file.\r\n",
 			    ebuf.buffer.buffer_len);
 		}
 		for (i = 0; i < ebuf.buffer.buffer_len; i ++) {
@@ -259,6 +348,115 @@ static int sendpkt(sb)
     return TRUE;
 }
 
+/*  GETAUDIO  --  Open audio input.  If audio hardware is half duplex,
+		  we may have to mute output in progress.  */
+
+static int getaudio()
+{
+    int i;
+
+#ifdef HALF_DUPLEX
+    struct soundbuf hdreq;
+    struct sockaddr_in name;
+
+    /* If we're afflicted with half-duplex hardware, make several
+       attempts to mute the output device and open audio input.
+       We send a mute even if the output isn't active as a courtesy
+       to the output program; if a later attempt to acquire output
+       fails, he'll know it was a result of being muted. */
+
+    for (i = 5; i > 0; i--) {
+	hdreq.compression = fHalfDuplex | fHalfDuplexMute;
+	strcpy(hdreq.sendinghost, hostname);
+	hdreq.buffer.buffer_len = 0;
+
+	bcopy((char *) &localhost, (char *) &(name.sin_addr), sizeof localhost);
+	name.sin_family = AF_INET;
+	name.sin_port = htons(Internet_Port);
+	if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL) +
+		   hdreq.buffer.buffer_len,
+		   0, (struct sockaddr *) &name, sizeof name) < 0) {
+            perror("sending half-duplex mute request");
+	    break;
+	}
+#ifdef BSD_like
+	 usleep(50000L);
+#endif
+#ifdef sgi
+	 sginap(50000L / usticks);
+#endif
+	if (soundinit(O_RDONLY)) {
+	    halfDuplexMuted = TRUE;
+	    break;
+	}
+    }
+
+    /* If we failed to initialise, send a resume just in case
+       one of our mute requests made it through anyway. */
+
+    if (i <= 0) {
+	hdreq.compression = fHalfDuplex | fHalfDuplexResume;
+	if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL) +
+		   hdreq.buffer.buffer_len,
+		   0, (struct sockaddr *) &name, sizeof name) < 0) {
+            perror("sending half-duplex resume request");
+	}
+    }
+#else
+    i = soundinit(O_RDONLY);
+#endif
+
+    if (i <= 0) {
+            fprintf(stderr, "%s: unable to initialise audio.\n", progname);
+	    return FALSE;
+    }
+    havesound = TRUE;
+    if (agc) {
+	soundrecgain(rgain);	  /* Set initial record level */
+    }
+    return TRUE;
+}
+
+/*  FREEAUDIO  --  Release sound input device, sending a resume to
+                   the output device if we've muted it due to
+		   half-duplex hardware.  */
+
+static void freeaudio()
+{
+    if (havesound) {
+	if (debugging) {
+            fprintf(stderr, "Restoring audio modes at release.\n");
+	}
+	soundterm();
+	havesound = FALSE;
+    }
+
+#ifdef HALF_DUPLEX
+
+    /* When bailing out, make sure we don't leave the output
+       muted. */
+
+    if (halfDuplexMuted) {
+	struct soundbuf hdreq;
+	struct sockaddr_in name;
+
+	hdreq.compression = fHalfDuplex | fHalfDuplexResume;
+	strcpy(hdreq.sendinghost, hostname);
+	hdreq.buffer.buffer_len = 0;
+
+	bcopy((char *) &localhost, (char *) &(name.sin_addr), sizeof localhost);
+	name.sin_family = AF_INET;
+	name.sin_port = htons(Internet_Port);
+	if (sendto(sock, &hdreq, (sizeof(struct soundbuf) - BUFL) +
+		   hdreq.buffer.buffer_len,
+		   0, (struct sockaddr *) &name, sizeof name) < 0) {
+            perror("sending half-duplex resume request");
+	}
+	halfDuplexMuted = FALSE;
+    }
+#endif
+}
+
 /*  EXITING  --  Catch as many program termination signals as
 		 possible and restore initial audio modes before
 		 we exit.  */
@@ -269,18 +467,13 @@ static void exiting()
     if (rawmode) {
         fprintf(stderr, "\r      \r");
 	fcntl(fileno(stdin), F_SETFL, 0);
-	noraw();
+	nocbreak();
 	echo();
 	endwin();
     }
 #endif
-    if (havesound) {
-	if (debugging) {
-            fprintf(stderr, "Restoring audio modes at exit.\n");
-	}
-	soundterm();
-	exit(0);
-    }
+    freeaudio();
+    exit(0);
 }
 
 /*  TERMCHAR  --  Check for special characters from console when
@@ -358,13 +551,13 @@ static int sendfile(f)
     /* Send a file */
 
     if (afile) {
-#ifdef sun
+#ifdef BSD_like
 	struct timeb t1, t2;
 #endif
 	long et;
 
 	while (
-#ifdef sun
+#ifdef BSD_like
 		    ftime(&t1),
 #endif
 		    (bread = nread =
@@ -414,7 +607,7 @@ static int sendfile(f)
 #define kOverhead   3000
 #endif
 
-#ifdef sun
+#ifdef BSD_like
 	    ftime(&t2);
 	    et = ((bread * 125) -
 		  1000 * (((t2.time - t1.time) * 1000) +
@@ -426,7 +619,7 @@ static int sendfile(f)
 #endif
 
 	    if (et > 0) {
-#ifdef sun
+#ifdef BSD_like
 		usleep(et);
 #endif
 #ifdef sgi
@@ -445,6 +638,8 @@ static int sendfile(f)
 	fclose(afile);
     } else {
 
+	    /* Send real-time audio. */
+
 #ifdef PUSH_TO_TALK
 	if (push) {
 	    char c;
@@ -454,7 +649,7 @@ static int sendfile(f)
 	    fflush(stderr);
 	    initscr();
 	    noecho();
-	    raw();
+	    cbreak();
 	    rawmode = TRUE;
 	    read(fileno(stdin), &c, 1);
 	    termchar(c);
@@ -466,142 +661,207 @@ static int sendfile(f)
 
 	/* Send real-time sound. */
 
-	if (!soundinit(O_RDONLY /* | O_NDELAY */ )) {
+	if (!getaudio()) {
             fprintf(stderr, "Unable to initialise audio.\n");
 	    return 2;
 	}
-	havesound = TRUE;
 	signal(SIGHUP, exiting);     /* Set signal to handle termination */
 	signal(SIGINT, exiting);     /* Set signal to handle termination */
 	signal(SIGTERM, exiting);    /* Set signal to handle termination */
-	if (agc) {
-	    soundrecgain(rgain);      /* Set initial record level */
-	}
-	if (soundrecord()) {
-	    while (TRUE) {
-		int soundel = soundgrab(buf,
-#ifdef TINY_PACKETS
-					     sound_packet
-#else
-					     sizeof buf
-#endif
-				       );
+	while (TRUE) {
+	    int soundel = 0;
+	    unsigned char *bs = (unsigned char *) buf;
 
-		unsigned char *bs = (unsigned char *) buf;
+	    if (havesound) {
+		soundel = soundgrab(buf,
+#ifdef TINY_PACKETS
+					 sound_packet
+#else
+					 sizeof buf
+#endif
+				   );
+	    }
 
 #ifdef PUSH_TO_TALK
-		if (push) {
-		    char c;
+	    if (push) {
+		char c;
 
-		    if (read(fileno(stdin), &c, 1) > 0) {
-			termchar(c);
-			talking = !talking;
-                        fprintf(stderr, "\r%s:  ", talking ? "Talk" : "Pause");
-			fflush(stderr);
-			fcntl(fileno(stdin), F_SETFL, talking ? O_NDELAY : 0);
-			if (talking) {
-			    /* Discard all backlog sound input. */
-			    soundflush();
-			}
-		    }
-/*
-		    if (!talking) {
-			continue;
-		    }
-*/
-		}
-#endif
+		if (read(fileno(stdin), &c, 1) > 0) {
+		    termchar(c);
+		    talking = !talking;
+		    fflush(stderr);
+#ifdef HALF_DUPLEX
 
-		if (soundel > 0) {
-		    register unsigned char *start = bs;
-		    register int j;
-		    int squelched = (squelch > 0), osl = soundel;
+		    /* For half-duplex, acquire and release the
+		       audio device at each transition.  This lets
+		       us mute the output only while in Talk mode. */
 
-		    /* If entire buffer is less than squelch, ditch it. */
-
-		    if (squelch > 0) {
-			for (j = 0; j < soundel; j++) {
-			    if (((*start++ & 0x7F) ^ 0x7F) > squelch) {
-				squelched = FALSE;
-				break;
-			    }
-			}
-		    }
-
-		    if (squelched) {
-			if (debugging) {
-                            printf("Entire buffer squelched.\n");
+		    if (talking) {
+			if (!getaudio()) {
+                            fprintf(stderr, "Audio device busy.\n");
+			    talking = FALSE;
 			}
 		    } else {
-			netbuf.compression = FALSE | (ring ? (fSetDest | fDestSpkr) : 0);
-			netbuf.compression |= debugging ? fDebug : 0;
-			netbuf.compression |= loopback ? fLoopBack : 0;
-
-			/* If automatic gain control is enabled,
-			   ride the gain pot to fill the dynamic range
-			   optimally. */
-
-			if (agc) {
-			    register unsigned char *start = bs;
-			    register int j;
-			    long msamp = 0;
-
-			    for (j = 0; j < soundel; j++) {
-				int s = audio_u2s(*start++);
-
-				msamp += (s < 0) ? -s : s;
-			    }
-			    msamp /= soundel;
-			    if (msamp < 6000) {
-				if (rgain < 100) {
-				    soundrecgain(++rgain);
-				}
-			    } else if (msamp > 7000) {
-				if (rgain > 1) {
-				    soundrecgain(--rgain);
-				}
-			    }
-			}
-
-			ring = FALSE;
-			if (compressing) {
-			    int i;
-
-			    soundel /= 2;
-			    for (i = 1; i < soundel; i++) {
-				buf[i] = buf[i * 2];
-			    }
-			    netbuf.compression |= fComp2X;
-			}
-			netbuf.buffer.buffer_len = soundel;
-			if (!sendpkt(&netbuf)) {
-			    return 1;
-			}
-			if (debugging) {
-                            fprintf(stderr, "Sent %d audio samples in %d bytes.\r\n",
-				    osl, netbuf.buffer.buffer_len);
-			}
+			freeaudio();
 		    }
-		} else {
-#ifdef sun
-		    usleep(100000L);  /* Wait for some sound to arrive */
 #endif
+                    fprintf(stderr, "\r%s:  ", talking ? "Talk" : "Pause");
+		    fcntl(fileno(stdin), F_SETFL, talking ? O_NDELAY : 0);
+		    if (talking) {
+			/* Discard all backlog sound input. */
+			soundflush();
+		    }
 		}
 	    }
-	} else {
-            fprintf(stderr, "Unable to start recording.\n");
-	    return 2;
+#endif
+
+	    if (soundel > 0) {
+		register unsigned char *start = bs;
+		register int j;
+		int squelched = (squelch > 0), osl = soundel;
+
+		/* If entire buffer is less than squelch, ditch it. */
+
+		if (squelch > 0) {
+		    for (j = 0; j < soundel; j++) {
+			if (((*start++ & 0x7F) ^ 0x7F) > squelch) {
+			    squelched = FALSE;
+			    break;
+			}
+		    }
+		}
+
+		if (squelched) {
+		    if (debugging) {
+                        printf("Entire buffer squelched.\n");
+		    }
+		} else {
+		    netbuf.compression = FALSE | (ring ? (fSetDest | fDestSpkr) : 0);
+		    netbuf.compression |= debugging ? fDebug : 0;
+		    netbuf.compression |= loopback ? fLoopBack : 0;
+
+		    /* If automatic gain control is enabled,
+		       ride the gain pot to fill the dynamic range
+		       optimally. */
+
+		    if (agc) {
+			register unsigned char *start = bs;
+			register int j;
+			long msamp = 0;
+
+			for (j = 0; j < soundel; j++) {
+			    int s = audio_u2s(*start++);
+
+			    msamp += (s < 0) ? -s : s;
+			}
+			msamp /= soundel;
+			if (msamp < 6000) {
+			    if (rgain < 100) {
+				soundrecgain(++rgain);
+			    }
+			} else if (msamp > 7000) {
+			    if (rgain > 1) {
+				soundrecgain(--rgain);
+			    }
+			}
+		    }
+
+		    ring = FALSE;
+		    if (compressing) {
+			int i;
+
+			soundel /= 2;
+			for (i = 1; i < soundel; i++) {
+			    buf[i] = buf[i * 2];
+			}
+			netbuf.compression |= fComp2X;
+		    }
+		    netbuf.buffer.buffer_len = soundel;
+		    if (!sendpkt(&netbuf)) {
+			return 1;
+		    }
+		    if (debugging) {
+                        fprintf(stderr, "Sent %d audio samples in %d bytes.\r\n",
+				osl, netbuf.buffer.buffer_len);
+		    }
+		}
+	    } else {
+#ifdef BSD_like
+		usleep(100000L);  /* Wait for some sound to arrive */
+#endif
+#ifdef sgi
+		sginap(100000L / usticks);
+#endif
+	    }
 	}
     }
     return 0;
 }
 
+/*  MAKESESSIONKEY  --	Generate session key with optional start
+			key.  If mode is TRUE, the key will be
+			translated to binary, otherwise it is
+			returned as 16 binary bytes.  */
+
+static void makeSessionKey(key, seed, mode)
+  char *key, *seed;
+  int mode;
+{
+    int j, k;
+    struct MD5Context md5c;
+    char md5key[16], md5key1[16];
+    char s[1024];
+
+    s[0] = 0;
+    if (seed != NULL) {
+	strcat(s, seed);
+    }
+    sprintf(s + strlen(s), "%u", getpid());
+    sprintf(s + strlen(s), "%u", getppid());
+    V getcwd(s + strlen(s), 256);
+    sprintf(s + strlen(s), "%u", clock());
+    V cuserid(s + strlen(s));
+    sprintf(s + strlen(s), "%u", time(NULL));
+    sprintf(s + strlen(s), "%u", gethostid());
+    getdomainname(s + strlen(s), 256);
+    gethostname(s + strlen(s), 256);
+    sprintf(s + strlen(s), "%u", getuid());
+    sprintf(s + strlen(s), "%u", getgid());
+    MD5Init(&md5c);
+    MD5Update(&md5c, s, strlen(s));
+    MD5Final(md5key, &md5c);
+    sprintf(s + strlen(s), "%u", (time(NULL) + 65121) ^ 0x375F);
+    MD5Init(&md5c);
+    MD5Update(&md5c, s, strlen(s));
+    MD5Final(md5key1, &md5c);
+    init_idearand(md5key, md5key1, time(NULL));
+    if (mode) {
+	for (j = k = 0; j < 16; j++) {
+	    unsigned char rb = idearand();
+
+#define Rad16(x) ((x) + 'A')
+	    key[k++] = Rad16((rb >> 4) & 0xF);
+	    key[k++] = Rad16(rb & 0xF);
+#undef Rad16
+	    if (j & 1) {
+                key[k++] = '-';
+	    }
+	}
+	key[--k] = 0;
+    } else {
+	for (j = 0; j < 16; j++) {
+	    key[j] = idearand();
+	}
+    }
+    close_idearand();
+}
+
 /*  USAGE  --  Print how-to-call information.  */
 
 static void usage()
 {
     V fprintf(stderr, "%s  --  Speak Freely sound sender.\n", progname);
-    V fprintf(stderr, "            Release 5.2, September 1995.\n");
+    V fprintf(stderr, "            Release 5.3, September 1995.\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "Usage: %s hostname [options] [ file1 / . ]...\n", progname);
     V fprintf(stderr, "Options: (* indicates defaults)\n");
@@ -619,13 +879,14 @@ static void usage()
     V fprintf(stderr, "           -L         Remote loopback\n");
     V fprintf(stderr, "     *     -M         Manual record gain control\n");
     V fprintf(stderr, "     *     -N         Do not compress subsequent sound\n");
-    V fprintf(stderr, "           -Ofile     Use file as one-time pad\n");
+    V fprintf(stderr, "           -Ofile     Use file as key file\n");
     V fprintf(stderr, "           -Phostname Party line, add host to list\n");
     V fprintf(stderr, "     *     -Q         Disable debug output\n");
     V fprintf(stderr, "           -R         Ring--force volume, output to speaker\n");
     V fprintf(stderr, "           -Sn        Squelch at level n (0-255)\n");
     V fprintf(stderr, "           -T         Telephone (GSM) compression\n");
     V fprintf(stderr, "           -U         Print this message\n");
+    V fprintf(stderr, "           -Z\"user..\" Send PGP session key for user(s)\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
     V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
@@ -646,12 +907,31 @@ main(argc, argv)
 
     progname = argv[0];
     gethostname(hostname, sizeof hostname);
+    if (strlen(hostname) > 15) {
+	hostname[15] = 0;
+    }
     gsmh = gsm_create();
 
 #ifdef sgi
     usticks = 1000000 / CLK_TCK;
 #endif
 
+#ifdef HALF_DUPLEX
+    {
+	struct hostent *h;
+	char host[512];
+
+	gethostname(host, sizeof host);
+	h = gethostbyname(host);
+	bcopy((char *) (h->h_addr), (char *) &localhost,
+	    sizeof localhost);
+#ifdef HDX_DEBUG
+        fprintf(stderr, "%s: local host %s = %s\n", progname, host,
+		inet_ntoa(localhost));
+#endif
+    }
+#endif
+
     /* Create the socket used to send data. */
 
     sock = socket(AF_INET, SOCK_DGRAM, 0);
@@ -695,38 +975,7 @@ main(argc, argv)
 		    break;
 
                 case 'E':             /* -E  --  Emit session key and exit */
-		    strcpy(s, op + 1);
-                    sprintf(s + strlen(s), "%u", getpid());
-                    sprintf(s + strlen(s), "%u", getppid());
-		    V getcwd(s + strlen(s), 256);
-                    sprintf(s + strlen(s), "%u", clock());
-		    V cuserid(s + strlen(s));
-                    sprintf(s + strlen(s), "%u", time(NULL));
-                    sprintf(s + strlen(s), "%u", gethostid());
-		    getdomainname(s + strlen(s), 256);
-		    gethostname(s + strlen(s), 256);
-                    sprintf(s + strlen(s), "%u", getuid());
-                    sprintf(s + strlen(s), "%u", getgid());
-		    MD5Init(&md5c);
-		    MD5Update(&md5c, s, strlen(s));
-		    MD5Final(md5key, &md5c);
-                    sprintf(s + strlen(s), "%u", (time(NULL) + 65121) ^ 0x375F);
-		    MD5Init(&md5c);
-		    MD5Update(&md5c, s, strlen(s));
-		    MD5Final(md5key1, &md5c);
-		    init_idearand(md5key, md5key1, time(NULL));
-		    for (j = k = 0; j < 16; j++) {
-			unsigned char rb = idearand();
-
-#define Rad32(x) ((x) + 'A')
-			s[k++] = Rad32((rb >> 4) & 0xF);
-			s[k++] = Rad32(rb & 0xF);
-			if (j & 1) {
-                            s[k++] = '-';
-			}
-		    }
-		    close_idearand();
-		    s[--k] = 0;
+		    makeSessionKey(s, op + 1, TRUE);
                     printf("%s\n", s);
 		    return 0;
 
@@ -797,23 +1046,23 @@ main(argc, argv)
 		    adpcmcompress = FALSE;
 		    break;
 
-                case 'O':             /* -Ofile -- Use file as one-time pad */
+                case 'O':             /* -Ofile -- Use file as key file */
 		    if (op[1] == 0) {
-			curotp = NULL; /* Switch off one-time pad */
+			curotp = NULL; /* Switch off key file */
 		    } else {
                         fp = fopen(op + 1, "r");
 			if (fp == NULL) {
-                            perror("Cannot open one-time pad file");
+                            perror("Cannot open key file");
 			    return 2;
 			}
 			curotp = malloc(BUFL);
 			if (curotp == NULL) {
-                            fprintf(stderr, "Cannot allocate one-time pad buffer.\n");
+                            fprintf(stderr, "Cannot allocate key file buffer.\n");
 			    return 2;
 			}
 			l = fread(curotp, 1, BUFL, fp);
 			if (l == 0) {
-                            /* Idiot supplied void one-time pad.  Give 'im
+                            /* Idiot supplied void key file.  Give 'im
 			       what he asked for: no encryption. */
 			    curotp[0] = 0;
 			    l = 1;
@@ -867,6 +1116,96 @@ main(argc, argv)
                 case '?':             /* -?  --  Print usage information */
 		    usage();
 		    return 0;
+
+                case 'Z':             /* -Z"user1 user2..."  --  Send PGP
+						 encrypted session key to
+						 named users */
+		    if (op[1] == 0) {
+			curpgpkey[0] = FALSE;
+		    } else {
+			char c[80], f[40];
+			FILE *kfile;
+			FILE *pipe;
+			long flen;
+
+                        sprintf(f, "/tmp/.SF_SKEY%d", getpid());
+                        sprintf(c, "pgp -fe +nomanual +verbose=0 %s >%s", op + 1, f);
+#ifdef PGP_DEBUG
+                        fprintf(stderr, "Encoding session key with: %s\n", c);
+#endif
+                        pipe = popen(c, "w");
+			if (pipe == NULL) {
+                            fprintf(stderr, "Unable to open pipe to: %s\n", c);
+			    return 2;
+			} else {
+			    makeSessionKey(curpgpkey + 1, NULL, FALSE);
+#ifdef PGP_DEBUG
+			    {	
+				int i;
+
+                                fprintf(stderr, "Session key:", hostname);
+				for (i = 0; i < 16; i++) {
+                                    fprintf(stderr, " %02X", curpgpkey[i + 1] & 0xFF);
+				}
+                                fprintf(stderr, "\n");
+			    }
+#endif
+			    /* The reason we start things off right with
+                               "Special K" is to prevent the session key
+			       (which can be any binary value) from
+                               triggering PGP's detection of the file as
+			       one already processed by PGP.  This causes an
+			       embarrassing question when we are attempting
+			       to run silent. */
+                            curpgpkey[0] = 'K';
+			    fwrite(curpgpkey, 17, 1, pipe);
+			    curpgpkey[0] = FALSE;
+			    fflush(pipe);
+			    pclose(pipe);
+			}
+                        kfile = fopen(f, "r");
+			if (kfile == NULL) {
+                            fprintf(stderr, "Cannot open key file %s\n", f);
+			} else {
+			    fseek(kfile, 0L, 2);
+			    flen = ftell(kfile);
+			    rewind(kfile);
+#ifdef PGP_DEBUG
+                            fprintf(stderr, "PGP key buffer length: %d\n", flen);
+#endif
+#ifdef TINY_PACKETS
+			    if (flen > (TINY_PACKETS -
+				    (sizeof(soundbuf) - BUFL))) {
+                                fprintf(stderr, "Warning: PGP key message exceeds %d packet size.\n", TINY_PACKETS);
+			    }
+#endif
+			    if (pgpsb != NULL) {
+				free(pgpsb);
+			    }
+			    pgpsb = (soundbuf *) malloc(((unsigned) flen) +
+				      (TINY_PACKETS - (sizeof(soundbuf) - BUFL)));
+			    if (pgpsb == NULL) {
+                                fprintf(stderr, "Cannot allocate PGP sound buffer.\n");
+				fclose(kfile);
+				unlink(f);
+				return 2;
+			    }
+			    fread(pgpsb->buffer.buffer_val, (int) flen, 1, kfile);
+			    pgpsb->buffer.buffer_len = flen;
+			    pgpsb->compression = fKeyPGP;
+#ifdef SENDMD5
+			    MD5Init(&md5c);
+			    MD5Update(&md5c, pgpsb->buffer.buffer_val, pgpsb->buffer.buffer_len);
+			    MD5Final(pgpsb->sendinghost, &md5c);
+#else
+			    strcpy(pgpsb->sendinghost, hostname);
+#endif
+			    fclose(kfile);
+			    curpgpkey[0] = TRUE;
+			}
+			unlink(f);
+		    }
+		    break;
 	    }
 	} else {
 	    if (dests == NULL) {
