@@ -6,7 +6,7 @@
 	Designed and implemented in July of 1991 by John Walker.
 
     Release 5: August 1995.  Added GSM compression, DES, IDEA, and
-			     one-time pad encryption.
+			     key file encryption.
 
     Release 5.1: September 1995.  Added ADPCM compression, connection
 			     logging (-v) and overrides for debug (-d)
@@ -29,17 +29,24 @@ static struct adpcm_state adpcm;      /* ADPCM compression state */
 static gsm gsmh;		      /* GSM handle */
 static char deskey[9] = "";           /* DES key, if any */
 static char ideakey[17] = "";         /* IDEA key, if any */
-static char *curotp = NULL;	      /* One-time pad, if any */
+static char *curotp = NULL;	      /* Key file, if any */
 struct sockaddr_in from;	      /* Sending host address */
 static int fromlen;		      /* Length of sending host address */
 static int showhosts = FALSE;	      /* Show host names that connect */
 static int hosttimeout = 60;	      /* Consider host idle after this time */
 static char *prog;		      /* Program name */ 
 static FILE *record = NULL;	      /* Answering machine record file */
+static char *pgppass = NULL;	      /* PGP secret key password */
+#ifdef HALF_DUPLEX
+static struct in_addr localhost;      /* Our internet address */
+static int halfDuplexMuted = FALSE;   /* Muted by half-duplex transmission */
+#endif
 
 struct connection {
     struct connection *con_next;      /* Next connection */
     struct in_addr con_addr;	      /* Host Internet address */
+    char pgpkey[17];		      /* PGP key for connection */
+    char keymd5[16];		      /* Digest of key file */
     int con_timeout;		      /* Connection timeout */
     char con_hostname[264];	      /* Host name */
 };
@@ -98,6 +105,21 @@ static void adpcmdecomp(sb)
     bcopy(dob, dp, sb->buffer.buffer_len);
 }
 
+/*  ETIME  --  Edit time and date for log messages.  */
+
+static char *etime()
+{
+    struct tm *t;
+    time_t clock;
+    static char s[20];
+
+    time(&clock);
+    t = localtime(&clock);
+    sprintf(s, "%02d-%02d %02d:%02d", t->tm_mon + 1, t->tm_mday,
+	       t->tm_hour, t->tm_min);
+    return s;
+}
+
 /*  RELEASE  --  Signal-catching function which releases the audio
                  device if we haven't received anything to play in
 		 the last minute. */
@@ -105,22 +127,36 @@ static void adpcmdecomp(sb)
 static void release()
 {
     struct connection *c, *l, *n;
+    int actives = 0;
 
     /* Mark idle any hosts that haven't sent us anything recently. */
 
     c = conn;
     l = NULL;
     while (c != NULL) {
-	c->con_timeout += TickTock;
+	if (c->con_timeout < hosttimeout) {
+	    c->con_timeout += TickTock;
+	    actives++;
+	}
 	n = c->con_next;
-	if (c->con_timeout >= hosttimeout) {
-            fprintf(stderr, "%s: %s idle\n", prog, c->con_hostname);
-	    if (l == NULL) {
-		conn = n;
-	    } else {
-		l->con_next = n;
+	if (c->con_timeout == hosttimeout) {
+	    actives--;		      /* Oops--just went inactive */
+	    c->con_timeout++;
+            fprintf(stderr, "%s: %s %s idle\n", prog, etime(), c->con_hostname);
+
+            /* If there's a PGP session key for this host, don't actually
+	       release the connection buffer, as that would lose it.  Issue
+	       an idle message, though, and tweak the timeout so we do this
+	       only once. */
+
+	    if (c->pgpkey[0] == 0) {
+		if (l == NULL) {
+		    conn = n;
+		} else {
+		    l->con_next = n;
+		}
+		free(c);
 	    }
-	    free(c);
 	} else {
 	    l = c;
 	}
@@ -148,7 +184,7 @@ static void release()
     /* If we still own the sound device or have a connection
        open, reset the timer. */
 
-    if (audiok || conn != NULL) {
+    if (audiok || (actives > 0)) {
 	alarm(TickTock);	      /* Wind the cat */
 	if (Debug) {
             fprintf(stderr, "%s: Tick....\n", prog);
@@ -162,8 +198,9 @@ static void release()
 
 /*  PLAYBUFFER	--  Send next buffer to audio output. */
 
-static void playbuffer(msg)
+static void playbuffer(msg, c)
   soundbuf *msg;
+  struct connection *c;
 {
     char *val;
     LONG len;
@@ -218,13 +255,27 @@ static void playbuffer(msg)
 	int i;
 
 	if (Debug) {
-            fprintf(stderr, "%s: decrypting %d bytes with one-time pad.\n", prog, len);
+            fprintf(stderr, "%s: decrypting %d bytes with key file.\n", prog, len);
 	}
 	for (i = 0; i < len; i ++) {
 	    val[i] ^= curotp[i];
 	}
     }
 
+    if ((msg->compression & fEncPGP) && c->pgpkey[0]) {
+	unsigned short iv[4];
+	LONG slen = (len + 7) & (~7);
+
+	bzero(iv, sizeof(iv));
+	initcfb_idea(iv, c->pgpkey + 1, TRUE);
+
+	if (Debug) {
+            fprintf(stderr, "%s: decrypting %d bytes with PGP key.\n", prog, slen);
+	}
+	ideacfb(val, slen);
+	close_idea();
+    }
+
     if ((msg->compression & fEncIDEA) && ideakey[0]) {
 	unsigned short iv[4];
 	LONG slen = (len + 7) & (~7);
@@ -306,18 +357,22 @@ static void playbuffer(msg)
 static void usage()
 {
     V fprintf(stderr, "%s  --  Speak Freely sound receiver.\n", prog);
-    V fprintf(stderr, "        Release 5.2, September 1995.\n");
+    V fprintf(stderr, "               Release 5.3, September 1995.\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "Usage: %s [options]\n", prog);
     V fprintf(stderr, "Options:\n");
     V fprintf(stderr, "           -D         Force debug output\n");
     V fprintf(stderr, "           -Ikey      IDEA decrypt with key\n");
     V fprintf(stderr, "           -Kkey      DES decrypt with key\n");
-    V fprintf(stderr, "           -Ofile     Use file as one-time pad\n");
+#ifdef MULTICAST
+    V fprintf(stderr, "           -Mhost/ip  Join multicast to given name or IP address\n");
+#endif
+    V fprintf(stderr, "           -Ofile     Use file as key file\n");
     V fprintf(stderr, "           -Q         Prevent debug output\n");
     V fprintf(stderr, "           -R[+]file  Record [append] sound in file\n");
     V fprintf(stderr, "           -U         Print this message\n");
     V fprintf(stderr, "           -Vtimeout  Show hostnames that connect\n");
+    V fprintf(stderr, "           -Z\"phrase\" Set PGP secret key pass phrase\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
     V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
@@ -334,14 +389,35 @@ main(argc, argv)
     struct sockaddr_in name;
     struct soundbuf sb;
     FILE *fp;
+    struct connection *c;
+    int newconn;
     struct auhdr {		      /* .au file header */
 	char magic[4];
 	long hsize, dsize, emode, rate, nchan;
     };
     static struct auhdr afh = { ".snd", sizeof(struct auhdr), 1, 1, 8000, 1 };
 
+    /* Create the socket from which to read */
+
+    sock = socket(AF_INET, SOCK_DGRAM, 0);
+    if (sock < 0) {
+        perror("opening datagram socket");
+	return 1;
+    }
+
+    /* Create name with wildcards. */
+
+    name.sin_family = AF_INET;
+    name.sin_addr.s_addr = INADDR_ANY;
+    name.sin_port = htons(Internet_Port);
+    if (bind(sock, (struct sockaddr *) &name, sizeof name) < 0) {
+        perror("binding datagram socket");
+	return 1;
+    }
+
     /*	Process command line options.  */
 
+    pgppass = getenv("PGPPASS");
     prog = argv[0];
     for (i = 1; i < argc; i++) {
 	char *op, opt;
@@ -391,20 +467,49 @@ main(argc, argv)
 		    }
 		    break;
 
-                case 'O':             /* -Ofile -- Use file as one-time pad */
+#ifdef MULTICAST
+                case 'M':             /* -Mhost/ip -- Join multicast to name/IP number */
+		    {
+			struct ip_mreq m;
+			struct hostent *h;
+			long naddr;
+
+                        /* If it's a valid IP number, use it.  Otherwise try to look
+			   up as a host name. */
+
+			if ((naddr = inet_addr(op + 1)) == -1) {
+			    h = gethostbyname(op + 1);
+			    if (h == 0) {
+                                fprintf(stderr, "%s: unknown multicast group\n", op + 1);
+				return 2;
+			    }
+			    bcopy((char *) h->h_addr, (char *) &naddr, sizeof naddr);
+			}
+			m.imr_multiaddr.s_addr = naddr;
+			m.imr_interface.s_addr = htons(INADDR_ANY);
+			if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
+			    &m, sizeof m) < 0) {
+                            perror("joining multicast group");
+			    return 2;
+			}
+		    }
+		    break;
+#endif
+
+                case 'O':             /* -Ofile -- Use file as key file */
                     fp = fopen(op + 1, "r");
 		    if (fp == NULL) {
-                        perror("Cannot open one-time pad file");
+                        perror("Cannot open key file");
 			return 2;
 		    }
 		    curotp = malloc(BUFL);
 		    if (curotp == NULL) {
-                        fprintf(stderr, "%s: Cannot allocate one-time pad buffer.\n", prog);
+                        fprintf(stderr, "%s: Cannot allocate key file buffer.\n", prog);
 			return 2;
 		    }
 		    l = fread(curotp, 1, BUFL, fp);
 		    if (l == 0) {
-                        /* Idiot supplied void one-time pad.  Give 'im
+                        /* Idiot supplied void key file.  Give 'im
 			   what he asked for: no encryption. */
 			curotp[0] = 0;
 			l = 1;
@@ -463,10 +568,33 @@ main(argc, argv)
 			    if (t < (TickTock + 1)) {
 				t = TickTock + 1;
 			    }
-			    hosttimeout = t;
+			    hosttimeout = (t / TickTock) * TickTock;
 			}
 		    }
 		    break;
+
+                case 'Z':             /* -Z"phrase"  --  Set PGP pass phrase */
+		    if (op[1] != 0) {
+			pgppass = strdup(op + 1);
+			op = argv[i];
+                        /* Zap the pass phrase in memory so "ps" etc. don't see it. */
+			while (*op != 0) {
+                            *op++ = ' ';
+			}
+		    } else {
+			char s[256];
+
+                        fprintf(stderr, "PGP secret key pass phrase: ");
+			fflush(stderr);
+			initscr();
+			noecho();
+			getstr(s);
+			pgppass = strdup(s);
+			echo();
+			endwin();
+                        fprintf(stderr, "\n");
+		    }
+		    break;
 	    }
 	} else {
 	    usage();
@@ -474,28 +602,28 @@ main(argc, argv)
 	}
     }
 
-    /* Create the socket from which to read */
+#ifdef HALF_DUPLEX
+    {
+	struct hostent *h;
+	char host[512];
 
-    sock = socket(AF_INET, SOCK_DGRAM, 0);
-    if (sock < 0) {
-        perror("opening datagram socket");
-	return 1;
+	gethostname(host, sizeof host);
+	h = gethostbyname(host);
+	bcopy((char *) (h->h_addr), (char *) &localhost,
+	    sizeof localhost);
+#ifdef HDX_DEBUG
+	if (Debug) {
+            fprintf(stderr, "%s: local host %s = %s\n", prog, host,
+		   inet_ntoa(localhost));
+	}
+#endif
     }
+#endif
 
     /* Initialise GSM decoding */
 
     gsmh = gsm_create();
 
-    /* Create name with wildcards. */
-
-    name.sin_family = AF_INET;
-    name.sin_addr.s_addr = INADDR_ANY;
-    name.sin_port = Internet_Port;
-    if (bind(sock, (struct sockaddr *) &name, sizeof name) < 0) {
-        perror("binding datagram socket");
-	return 1;
-    }
-
     /* Find assigned port value and print it. */
 
     length = sizeof(name);
@@ -533,6 +661,43 @@ main(argc, argv)
 	}
 #endif
 
+#ifdef HALF_DUPLEX
+
+	/* If this is a half duplex mute request, immediately release
+	   the audio output device if we have it.  We verify the
+           sender's address and accept mute requests only from the
+	   local host; nobody else has any business telling us
+	   to shut up! */
+
+	if (sb.compression & fHalfDuplex) {
+#ifdef HDX_DEBUG
+	    if (Debug) {
+                static char *hdxreq[4] = { "Bogus(0)", "Mute", "Resume",
+                                           "Bogus(3)" };
+		unsigned char *cp = (unsigned char *) &localhost;
+
+                fprintf(stderr, "%s: half-duplex %s request from %s.\n",
+		    prog, hdxreq[sb.compression & 3], sb.sendinghost);
+	    }
+#endif
+	    if (memcmp(&from.sin_addr, &localhost, sizeof localhost) == 0) {
+		if (sb.compression & fHalfDuplexMute) {
+		    if (audiok) {
+			soundterm();
+			audiok = FALSE;
+			if (Debug) {
+                            fprintf(stderr, "%s: half-duplex releasing audio output.\n", prog);
+			}
+		    }
+		    halfDuplexMuted = TRUE;
+		} else if (sb.compression & fHalfDuplexResume) {
+		    halfDuplexMuted = FALSE;
+		}
+	    }
+	    continue;		      /* Done with packet */
+	}
+#endif
+
 	/* If the packet requests loop-back, immediately dispatch it
 	   back to the host who sent it to us.	To prevent an infinite
 	   loop-back cycle, we clear the loop-back bit in the header
@@ -551,44 +716,139 @@ main(argc, argv)
 	   }
 	}
 
-	if (showhosts) {
-	    struct connection *c = conn;
+	/* See if this connection is active.  If not, initialise a new
+	   connection. */
 
-	    while (c != NULL) {
-		if (memcmp(&from.sin_addr, &(c->con_addr),
-			   sizeof(struct in_addr)) == 0) {
-		    break;
-		}
-		c = c->con_next;
-	    }
-	    if (c == NULL) {
-		c = (struct connection *) malloc(sizeof(struct connection));
-		if (c != NULL) {
-		    struct hostent *h;
-
-		    c->con_next = conn;
-		    conn = c;
-		    bcopy(&from.sin_addr, &(c->con_addr),
-			sizeof(struct in_addr));
-		    h = gethostbyaddr((char *) &from.sin_addr, sizeof(struct in_addr),
-				      AF_INET);
-		    if (h == NULL) {
-			unsigned char *ip = (unsigned char *) &from.sin_addr;
-                        sprintf(c->con_hostname, "%d.%d.%d.%d",
-			    ip[0], ip[1], ip[2], ip[3]);
-		    } else {
-			strcpy(c->con_hostname, h->h_name);
-		    }
-                    fprintf(stderr, "%s: %s connect\n", prog, c->con_hostname);
-		}
+	newconn = FALSE;
+	c = conn;
+	while (c != NULL) {
+	    if (memcmp(&from.sin_addr, &(c->con_addr),
+		       sizeof(struct in_addr)) == 0) {
+		break;
 	    }
+	    c = c->con_next;
+	}
+	if (c == NULL) {
+	    c = (struct connection *) malloc(sizeof(struct connection));
 	    if (c != NULL) {
-		/* Reset connection timeout. */
-		c->con_timeout = 0;
+		struct hostent *h;
+
+		newconn = TRUE;
+		c->con_next = conn;
+		c->pgpkey[0] = FALSE;
+		bzero(c->keymd5, 16);
+		conn = c;
+		bcopy(&from.sin_addr, &(c->con_addr),
+		    sizeof(struct in_addr));
+		h = gethostbyaddr((char *) &from.sin_addr, sizeof(struct in_addr),
+				  AF_INET);
+		if (h == NULL) {
+		    strcpy(c->con_hostname, inet_ntoa(from.sin_addr));
+		} else {
+		    strcpy(c->con_hostname, h->h_name);
+		}
+	    }
+	} else if (c->con_timeout == (hosttimeout + 1)) {
+	    newconn = TRUE;
+	}
+	if (c != NULL) {
+	    /* Reset connection timeout. */
+	    c->con_timeout = 0;
+	    if (newconn) {
+                fprintf(stderr, "%s: %s %s connect\n", prog, etime(), c->con_hostname);
 	    }
 	}
 
-	playbuffer(&sb);
+#ifdef HALF_DUPLEX
+
+        /* If we're muted by a transmission in progress on half-duplex
+	   audio hardware, this is the end of line for this sound
+	   packet. */
+
+	if (halfDuplexMuted) {
+	    if (Debug) {
+                fprintf(stderr, "%s: %s packet lost by half-duplex muting.\n",
+				prog, c->con_hostname);
+	    }
+	    continue;
+	}
+#endif
+
+	if ((sb.compression & fKeyPGP)) {
+	    char cmd[256], f[40], kmd[16];
+	    FILE *kfile;
+	    FILE *pipe;
+	    long flen;
+	    struct MD5Context md5c;
+
+	    MD5Init(&md5c);
+	    MD5Update(&md5c, sb.buffer.buffer_val, sb.buffer.buffer_len);
+	    MD5Final(kmd, &md5c);
+
+	    if (memcmp(c->keymd5, kmd, 16) != 0) {
+		bcopy(kmd, c->keymd5, 16);
+                sprintf(f, "/tmp/.SF_SKEY%d", getpid());
+
+                kfile = fopen(f, "w");
+		if (kfile == NULL) {
+                    fprintf(stderr, "Cannot open encrypted session key file %s\n", f);
+		} else {
+		    fwrite(sb.buffer.buffer_val, sb.buffer.buffer_len, 1, kfile);
+		    fclose(kfile);
+#ifdef ZZZ
+		    if (pgppass == NULL) {
+			static char s[256]; 
+
+                        fprintf(stderr, "Enter PGP pass phrase: ");
+			if (fgets(s, sizeof s, stdin) != NULL) {
+			    s[strlen(s) - 1] = 0;
+			    pgppass = s;
+			}
+		    }
+#endif
+                    sprintf(cmd, "pgp -f +nomanual +verbose=0 %s%s%s <%s",
+                        pgppass ? "-z\"" : "", pgppass ? pgppass : "",
+                        pgppass ? "\" " : "", f);
+#ifdef PGP_DEBUG
+                    fprintf(stderr, "Decoding session key with: %s\n", cmd);
+#else
+		    if (Debug) {
+                       fprintf(stderr, "%s: decoding PGP session key.\n", prog);
+		    }
+#endif
+                    pipe = popen(cmd, "r");
+		    if (pipe == NULL) {
+                        fprintf(stderr, "Unable to open pipe to: %s\n", c);
+		    } else {
+			if (fread(c->pgpkey, 1, 17, pipe) == 17) {
+			    c->pgpkey[0] = TRUE;
+#ifdef PGP_DEBUG
+			    {	
+				int i;
+
+                                fprintf(stderr, "Session key for %s:", c->con_hostname);
+				for (i = 0; i < 16; i++) {
+                                    fprintf(stderr, " %02X", c->pgpkey[i + 1] & 0xFF);
+				}
+                                fprintf(stderr, "\n");
+			    }
+#else
+			    if (Debug) {
+                               fprintf(stderr, "%s: PGP session key decoded.\n", prog);
+			    }
+#endif
+			} else {
+			    c->pgpkey[0] = FALSE;
+                            fprintf(stderr, "%s: Error decoding PGP session key.\n", prog);
+			}
+			pclose(pipe);
+		    }
+		    unlink(f);
+		}
+	    }
+	} else {
+	    playbuffer(&sb, c);
+	}
     }
 #ifdef MEANS_OF_EXIT
     close(sock);
